#include "common.h"
#include "tp_stub.h"
#include <algorithm>

#define OPTION_COUNT 32
static bool layerExSaveLoaded = false;
extern iTVPFunctionExporter * TVPFunctionExporter;
static std::wstring xp3path;

/**
* @filename should like a.tlg
* @path should like file:///xx .png
**/ 
void SaveTLGToPng(const wchar_t * filename, const wchar_t * path);

// global vars
int closed = 0;
void (* print_log) (const wchar_t*, int);
void (* print_log_screen) (const wchar_t*, int);
void (* update_last_info) (unsigned int, unsigned int);

//unpack a arc
static unsigned char spcode[] = {0x8B, 0xF0, 0x8B, 0xC6, 0x83, 0xE0, 0x07, 0x85, 0xC0};
static unsigned char edcode[] = {0x5E, 0x5B, 0x8B, 0xE5, 0x5D, 0xC3};

static unsigned char *guess_dummy_kmp(unsigned char* start, unsigned char* end, unsigned char* stand, int len)  {
	unsigned char* u = start;
	int state = 0;
	for (; u < (unsigned char *)end; ++u) {
		if (*u == stand[state]) ++state; else {
			state = 0;
			if (stand[0] == *u) state = 1;
		}
		if (state == len) { //attached!
			return u - len;
		}
	}
	return 0;
}

static unsigned char func_start[] = {0x55, 0x8B, 0xEC};
unsigned int search_delta(char * point_now, unsigned int * pstart) {
	static unsigned char deltacode[] = {0x53, 0x56, 0x57, 0x89, 0x95};
	unsigned int cnt = 0;
	while (memcmp(point_now, deltacode, sizeof(deltacode)) != 0) {
		--point_now;
		if (++cnt >= 0x1500) {
			return 0;
		}
	}
	unsigned int * pt = (unsigned int *) (point_now + 5);
	cnt = 0;
	while (memcmp(point_now, func_start, sizeof(func_start)) != 0) {
		--point_now;
		if (++cnt >= 0x20) {
			return 0;
		}
	}
	*pstart = (unsigned int) point_now ;
	return *pt;
}

static unsigned int search_cmd_place(unsigned char* begin, unsigned char* end, unsigned int addr) {
	while (begin != end) {
		if (*begin == 0xE8 && (*(unsigned int*)(begin+1) == addr - (unsigned int)begin - 5)) {
			unsigned int cnt = 0;
			while (memcmp(begin, func_start, sizeof(func_start)) != 0) {
				--begin;
				if (++cnt >= 0x100) {
					return 0;
				}
			}
			return (unsigned int)begin;
		}
		++begin;
	}
	return 0;
}

static int read_file ( func_info* info ) {
	FILE * fp = fopen("address.info.txt", "r");
	if (fp == NULL) {
		return 0;
	}
	fscanf(fp,"%d %d", &info->arc_construct_begin, &info->argument_delta);
	fclose(fp);
	return 1;
}

int seach_infomation(func_info* info) {
	unsigned char *place;
	if (read_file( info )) {
		place = (unsigned char * )info->arc_construct_begin + 0x10;
	} else {
		place = guess_dummy_kmp((unsigned char*)0x00401000, (unsigned  char*)0x00700000, spcode, 9);
		if (place == 0) {
			print_log(L"Failed search the construct function. Please send the information to author.");
			return -1;
		}
		info->argument_delta = search_delta((char*) place, &(info->arc_construct_begin));
	}

	if (place == 0) {
		print_log(L"Failed search the construct function. Please send the information to author.");
		return -1;
	}
	
	
	if (info->argument_delta == 0) {
		print_log(L"delta_search_failed");
		return -1;
	} 
	
	if (!(place = guess_dummy_kmp(place+1, (unsigned  char*)0x00700000, edcode, 6))) {
		print_log(L"xp3 construct hook miss");
		return -1;
	}
	
	info->open_archive = search_cmd_place((unsigned char*) 0x401000, (unsigned  char*)0x700000, info->arc_construct_begin);
	if (!info->open_archive) {
		print_log(L"open archive hook miss");
		return -1;
	}
	
	info->arc_construct_end = (unsigned int) place + 6;
	
	return 0;
}	


using namespace std;

wstring int2str(int number) {
	wstring res;
	wchar_t ch = L' ';
	if (number < 0) {
		ch = L'-';
		number = - number;
	}
	while (number) {
		res += wchar_t(number % 10 + 48);
		number /= 10;
	}
	reverse(res.begin(), res.end());	
	
	return ch + res;
}

wstring hex2str(unsigned int number) {
	wstring res;
	while (number) {
		int now = number % 16;
		if (now < 10) {
			res += wchar_t(now + 48);
		} else {
			res += wchar_t(now + 'a' - 10);
		}
		number /= 16;
	}

	reverse(res.begin(), res.end());
	return res;
}


// ==== dump part ====

static unsigned int __declspec(naked) get_count(unsigned int hdl) {
	__asm {
		mov eax, hdl
		mov ecx, [eax]
		jmp [ecx+4]
	}
}

static int make_sure_path(const wchar_t * _path) {
	if (_path[2] != '/' && _path[2] != '\\') {
		return 0;
	}
	int i = 3, done = 0, len = 0;
	while (_path[len]) ++len;

	wchar_t * path = new wchar_t[len+1];
	memcpy(path, _path, (len + 1)* 2);

	while (!done) {
		while (path[i] != '/' && path[i] != '\\' && path[i] != 0) {
			++i;
		}
		if (path[i] == 0) {
			done = true;
		} else {
			wchar_t last = path[i];
			path[i] = 0;
			HANDLE hFind;
			WIN32_FIND_DATA fdata;
			hFind = FindFirstFile(path, &fdata);
			if (hFind == INVALID_HANDLE_VALUE) {
				BOOL flag = CreateDirectory(path, NULL);
				if (flag == FALSE) {
					print_log(path);
					delete path;
					return 0;
				}
			} else {
				CloseHandle(hFind);
			}
			path[i++] = last;
		}
	}

	//succeed
	delete [] path;
	return 1;
}

static char buf[100000];
void dump_file(unsigned int handle, unsigned int id, const wchar_t* path) {
	ttstr fname;

	__asm {
		mov eax, handle
		mov ebx, [eax]
		lea ecx, fname
		mov edx, id
		call [ebx+8]
	}
	print_log(fname.c_str(), 0);
	print_log(L" : ", 0);

	if (fname.GetLen() > 250) {
		print_log(L"File name too long. ignored");
		return ;
	}
	
	wstring dir(path);
	dir += L'/';
	dir += fname.c_str();
	if (!make_sure_path(dir.c_str())) {
		print_log(L" failed when build path" );
		return ;
	}

	if (GetOption(OPT_TLG_TO_PNG)) {
		int len = dir.size();
		if (len > 3 && wstring(dir.end()-4) ==  L".tlg") {
			dir.end()[-3]='p'; dir.end()[-2]='n'; dir.end()[-1]='g';
			SaveTLGToPng(fname.c_str(), format_path(dir.c_str()).c_str());
			print_log(L"saved png");
			return;
		}
	}

	tTJSBinaryStream* _st;
	// create stream
	__asm {
		mov eax, handle
		mov ecx, [eax]
		mov edx, id
		call [ecx + 0Ch]
		mov _st, eax
	}

	if (!_st) {
		print_log(L" Failed when create Stream ");
		return ;
	}

	HANDLE hFile = CreateFile( dir.c_str(), GENERIC_WRITE, FILE_SHARE_WRITE,
		NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hFile == INVALID_HANDLE_VALUE) {
		print_log(L" Failed when create file ");
		return ;
	}

	unsigned fsize = 0, stream_size;
	__asm {
		mov eax, _st 
		mov ecx, [eax]
		push eax
		call [ecx+010h]
		mov stream_size, eax
	}

	while (1) {
		int tsize;
		__asm {
			push 65536D
			lea eax, buf
			push eax
			mov eax, _st
			mov ecx, [eax]
			push eax
			call [ecx+4]
			mov tsize, eax
		}
		if (!tsize) break;
		unsigned int now = 0, res;
		while (1) {
			WriteFile(hFile, buf+now, tsize-now, (LPDWORD)&res, NULL);
			tsize -= res;
			fsize += res;
			if (res == 0) break;
		}
		if (tsize) break;
	}

	CloseHandle(hFile);
	if (fsize != stream_size) {
		print_log(L"some error happened when dumping.");
	} else {
		print_log(L"finished");
	}
}

void unpack_archive(unsigned int handle, const wchar_t * path) {
	unsigned int count = get_count(handle), i;
	print_log((L"total files : " + int2str(count)).c_str());

	if (!make_sure_path((wstring(path) + L"\\a").c_str())) {
		print_log((L" failed when build path : " + wstring(path)).c_str());
		return ;
	}
	for (i = 0; i < count; ++i) {
		dump_file(handle, i, path);
		update_last_info(i+1, count);
		if (closed) {
			break;
		} 
	}
	print_log(L"");
	print_log_screen(L"");
}

// dump path
static wstring dump_path;

void set_dump_path(const wchar_t* s) {
	if  (s != NULL) {
		dump_path = s;
		while (dump_path[dump_path.size()] == L'/' || dump_path[dump_path.size()] == L'\\') {
			dump_path.erase(dump_path.size() - 1, 1);
		}
	}
}

std::wstring get_dump_path() {
	return dump_path;
}

// format path to url
std::wstring format_path(const wchar_t * path) {
	if (memcmp(path, L"file:", 10)==0) {
		return path;
	} else if (path[1] == L':' && ( path[0] <= L'Z' && path[0] >= L'A' ||  path[0] <=L'z' && path[0] >= L'a')) {
		wstring res(L"file://./");
		res += path[0];
		unsigned int i;
		for (i = 2; path[i]!=0; ++i) {
			if (path[i] == '\\')
				res += '/';
			else
				res += path[i];
		}
		return res;
	} else {
		unsigned int i, flag = 0;
		for (i = 0; path[i] != 0; ++i) {
			if (path[i] == '/' || path[i] =='\\' || path[i] =='*') {
				flag = 1;
				break;
			}
		}
		if (!flag) {
			wchar_t * buffer = new wchar_t [512];
			GetCurrentDirectory( 512 , buffer );
			wstring res = format_path((wstring(buffer) + L'/' + path).c_str());
			delete [] buffer;
			return res;
		}
	}

	print_log((wstring(L"failed change path format :") + path).c_str());
	print_log_screen((wstring(L"failed change path format :") + path).c_str());
	return L"";
}


// Options
static int options[OPTION_COUNT];
void SetOption(int item, int value) {
	if (item >= 0 && item <OPTION_COUNT)
		options[item] = value;
}

int GetOption(int item) {
	return (item >= 0 && item <OPTION_COUNT) ?
		options[item] : 0;
}

// tlg to png

static void loadPngDll() {
	HMODULE hModule = LoadLibrary(TEXT("layerExSave.dll"));
	if (hModule == NULL) {
		MessageBoxW(GetDesktopWindow(), TEXT("Failed Load layerExSave.dll"), TEXT("error"), 0);
		exit(0);
	}

	typedef HRESULT (_stdcall *tlink)(iTVPFunctionExporter *);
	tlink v2link = (tlink)GetProcAddress(hModule, "V2Link");
	wstring buf = L"v2link = " + hex2str(DWORD(v2link)) + L"  FuncPtr = " + hex2str(DWORD(TVPFunctionExporter));
	print_log(buf.c_str());
	v2link(TVPFunctionExporter);
	layerExSaveLoaded = true;
	print_log(L"InitOver");
}

static void PngInitialize() {
	TVPExecuteScript(ttstr(L"var tmp=new Window();"));
    TVPExecuteScript(ttstr(L"var tmpLayer=new Layer(tmp, null);"));
    TVPExecuteScript(ttstr(L"tmpLayer.opacity=255;"));
}

void PngAddXp3(const wchar_t * name) {
	xp3path = format_path(name);
	wstring str = L"Storages.addAutoPath(\"";
		str += xp3path;
		str += L">\");";
	xp3path += L'>';
	TVPExecuteScript(ttstr(str.c_str()));
}

wstring ConstU = L"tmpLayer.saveLayerImagePng(\"";

void SaveTLGToPng(const wchar_t * filename, const wchar_t * path) {
	if (!layerExSaveLoaded) {
		loadPngDll();
		PngInitialize();
	}
	wstring str, fname = filename;
	str = L"tmpLayer.loadImages(\"" + xp3path + fname + L"\");";
	TVPExecuteScript(ttstr(str.c_str()));
    TVPExecuteScript(ttstr(L"tmpLayer.setSizeToImageSize();"));
	str = ConstU + wstring(path) + L"\");";
	TVPExecuteScript(ttstr(str.c_str()));
}
